/*/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  VC8E-WFD V 0.5

  VC8E wireless display/controller for X-Y display of PDP/8 Spacewar game

  Designed to work with hardware version 1.0.

  This Arduino code for the ESP8266 can be compiled by the amazing and fabulous development
  tools/environment described at and downloadable from:

  https://www.arduino.cc/
  https://github.com/esp8266/Arduino

  A library that may not be included in the "standard distribution" needed, WiFiManater, is at:

  https://github.com/tzapu/WiFiManager

  The libraries are standard and by and large open source EXCEPT for the PFOD Buffered TCP
  client library at:

  http://www.forward.com.au/pfod/pfodParserLibraries/index.html

  It is copyrighted, but free to use personally/commercially. If this is a problem, there are
  other libraries of this type.

  This project was inspired by the oscilloscope clock project (I have two of these) at:

  http://www.dutchtronix.com/ScopeClock.htm

  And makes use of the 8x8 font described in the source code there, which is also open source.

  Ring buffer code bummed from the (very nice) Internet Radio project at:

  https://github.com/Edzelf/Esp-radio

  Additional thanks to the nameless computing professionals, geeks, dorks, wizards, twits,
  hackers, and software gods/godesses whose code and ideas I bummed/was inspired by on the collective Internet.

  _DO NOT_ even consider any of this code "good", or as an example of how to write in C,
  particularly in the use of defines. I have, however, reduced the number of "magic numbers"
  to a minimum. 

  My personal code here is released on the YCLAMC licence, which is, You Cannot Laugh
  At My Code, in addition to I hadn't seriously written in this language in the 21st century, I'm "old",
  and I was never really good at it anyway! But the code does indeed work; constructive
  improvements/suggestions, are always welcomed. I do intend to make the following improvements:

  1. Less flashing (you can tweek the timing numbers for beam on/off delays)
  2. Better recovery from connect failures and setup (when in doubt, reboot (power on/off, even after saving parameters)
  3. Use of the serial port
  4. Connect/disconnect status via the scope screen

  Information on the ESP8266 System on a Chip (SoC), which I consider to be one of the most
  important chips of the 21st century so far, is available at http://www.esp8266.com.
  Modules are available "everywhere"; if you can wait, they are available at somewhat lower
  cost on Ebay from China (buy several, you _will_ use them). If you have been doing this
  for...a while... you will not believe the power of this low-cost device, and I don't
  believe that the the SDK and tools are entirely optimized...yet. An even better version
  is on the way.

*///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////





#include "ESP8266WiFi.h"
#include <DNSServer.h>            //Local DNS Server used for redirecting all requests to the configuration portal
#include <ESP8266WebServer.h>     //Local WebServer used to serve the configuration portal
#include <WiFiManager.h>
#include <ESP8266mDNS.h>
#include <WiFiClient.h>
#include <EEPROM.h>
#include <pfodESP8266BufferedClient.h>
#include "eagle_soc.h"

#define THE_VERSION "V.0.5"
#define EEPROM_USE 512
#define CHAR_MAG_FACTOR 3 //this is a shift left
#define ONE_MILLI 4166
#define WEDGE_PHO 128
#define WEDGE_R 64
#define WEDGE_L 32
#define WEDGE_THR 16
#define WEDGE_HYP (WEDGE_R | WEDGE_L)
#define CALIB WEDGE_THR
#define BOOTIT  WEDGE_HYP
#define SHOW_SETUP WEDGE_PHO

#define WIFI_AP_NET "VC8E-WFD"
#define WIFI_AP_PASS "DECPDP8I"//don't make <8, firm/libs dont like


#define NEEDLE_PHO 8
#define NEEDLE_R 4
#define NEEDLE_L 2
#define NEEDLE_THR 1
#define NEEDLE_HYP (NEEDLE_R | NEEDLE_L)


#define HC165_CE 14   //"Normal" assignments of GPIO pins
#define HC165_CP 2
#define HC165_PL 15
#define HC165_NDATA 13
#define BEAMPIN 12
#define TLV5617_CS 5
#define TLV5617_CLK 2
#define TLV5617_DTA 15

#define TLV5617_CS_FAST _BV(TLV5617_CS)   //"Fast" assignments (for write ONLY)
#define TLV5617_CLK_FAST _BV(TLV5617_CLK)
#define TLV5617_DTA_FAST _BV(TLV5617_DTA)
#define BEAMPIN_FAST _BV(BEAMPIN)
#define HC165_CE_FAST _BV(HC165_CE)
#define HC165_CP_FAST _BV(HC165_CP)
#define HC165_PL_FAST _BV(HC165_PL)
#define HC165_NDATA_FAST _BV(HC165_NDATA)

#define BEAM_TIMELENGTH ONE_MILLI*5  //approx usec beam on
#define LOOPS_TO_PADDLES 300  //how many dots to look at the paddles


// Character info, from Scope Clock Project.

char  fdata[]  = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,    //32 (space)
  0x10, 0x38, 0x38, 0x10, 0x10, 0x00, 0x10, 0x00,
  0x6C, 0x6C, 0x48, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x28, 0x7C, 0x28, 0x28, 0x7C, 0x28, 0x00,
  0x20, 0x38, 0x40, 0x30, 0x08, 0x70, 0x10, 0x00,
  0x64, 0x64, 0x08, 0x10, 0x20, 0x4C, 0x4C, 0x00,
  0x20, 0x50, 0x50, 0x20, 0x54, 0x48, 0x34, 0x00,
  0x30, 0x30, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x10, 0x20, 0x20, 0x20, 0x20, 0x20, 0x10, 0x00, //40,
  0x20, 0x10, 0x10, 0x10, 0x10, 0x10, 0x20, 0x00,
  0x00, 0x28, 0x38, 0x7C, 0x38, 0x28, 0x00, 0x00,
  0x00, 0x10, 0x10, 0x7C, 0x10, 0x10, 0x00, 0x00,
  //    0x00,0x00,0x00,0x00,0x00,0x30,0x30,0x20,    //44 ',' (OLD)
  0x00, 0x00, 0x00, 0x00, 0x00, 0x60, 0x60, 0x40, //44 ','
  //    0x00,0x00,0x00,0x7C,0x00,0x00,0x00,0x00,    //45 '-' (OLD)
  0x00, 0x00, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, //45 '-'
  0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x30, 0x00,
  0x00, 0x04, 0x08, 0x10, 0x20, 0x40, 0x00, 0x00,
  0x38, 0x44, 0x4C, 0x54, 0x64, 0x44, 0x38, 0x00, //48 '0'
  //    0x38,0x44,0x44,0x44,0x44,0x44,0x38,0x00,    //48 '0'
  0x10, 0x30, 0x10, 0x10, 0x10, 0x10, 0x38, 0x00,
  0x38, 0x44, 0x04, 0x18, 0x20, 0x40, 0x7C, 0x00, //50,
  0x38, 0x44, 0x04, 0x38, 0x04, 0x44, 0x38, 0x00,
  0x08, 0x18, 0x28, 0x48, 0x7C, 0x08, 0x08, 0x00,
  0x7C, 0x40, 0x40, 0x78, 0x04, 0x44, 0x38, 0x00,
  0x18, 0x20, 0x40, 0x78, 0x44, 0x44, 0x38, 0x00,
  0x7C, 0x04, 0x08, 0x10, 0x20, 0x20, 0x20, 0x00,
  0x38, 0x44, 0x44, 0x38, 0x44, 0x44, 0x38, 0x00,
  0x38, 0x44, 0x44, 0x3C, 0x04, 0x08, 0x30, 0x00,
  //    0x00,0x00,0x30,0x30,0x00,0x30,0x30,0x00,    //58 ':' (OLD)
  0x00, 0x00, 0x60, 0x60, 0x00, 0x60, 0x60, 0x00, //58 ':'
  0x00, 0x00, 0x30, 0x30, 0x00, 0x30, 0x30, 0x20,
  0x08, 0x10, 0x20, 0x40, 0x20, 0x10, 0x08, 0x00, //60,
  0x00, 0x00, 0x7C, 0x00, 0x00, 0x7C, 0x00, 0x00,
  0x20, 0x10, 0x08, 0x04, 0x08, 0x10, 0x20, 0x00,
  0x38, 0x44, 0x04, 0x18, 0x10, 0x00, 0x10, 0x00,
  0x38, 0x44, 0x5C, 0x54, 0x5C, 0x40, 0x38, 0x00,
  0x38, 0x44, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x00,
  0x78, 0x44, 0x44, 0x78, 0x44, 0x44, 0x78, 0x00,
  0x38, 0x44, 0x40, 0x40, 0x40, 0x44, 0x38, 0x00, //67
  0x78, 0x44, 0x44, 0x44, 0x44, 0x44, 0x78, 0x00,
  0x7C, 0x40, 0x40, 0x78, 0x40, 0x40, 0x7C, 0x00,
  0x7C, 0x40, 0x40, 0x78, 0x40, 0x40, 0x40, 0x00, //70,
  0x38, 0x44, 0x40, 0x5C, 0x44, 0x44, 0x3C, 0x00,
  0x44, 0x44, 0x44, 0x7C, 0x44, 0x44, 0x44, 0x00,
  0x38, 0x10, 0x10, 0x10, 0x10, 0x10, 0x38, 0x00,
  0x04, 0x04, 0x04, 0x04, 0x44, 0x44, 0x38, 0x00,
  0x44, 0x48, 0x50, 0x60, 0x50, 0x48, 0x44, 0x00,
  0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x7C, 0x00,
  0x44, 0x6C, 0x54, 0x44, 0x44, 0x44, 0x44, 0x00,
  0x44, 0x64, 0x54, 0x4C, 0x44, 0x44, 0x44, 0x00,
  0x38, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x00, //79  'O'
  0x78, 0x44, 0x44, 0x78, 0x40, 0x40, 0x40, 0x00, //80,
  0x38, 0x44, 0x44, 0x44, 0x54, 0x48, 0x34, 0x00,
  0x78, 0x44, 0x44, 0x78, 0x48, 0x44, 0x44, 0x00,
  0x38, 0x44, 0x40, 0x38, 0x04, 0x44, 0x38, 0x00,
  0x7C, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x00,
  0x44, 0x44, 0x44, 0x44, 0x44, 0x44, 0x38, 0x00,
  0x44, 0x44, 0x44, 0x44, 0x44, 0x28, 0x10, 0x00,
  0x44, 0x44, 0x54, 0x54, 0x54, 0x54, 0x28, 0x00,
  0x44, 0x44, 0x28, 0x10, 0x28, 0x44, 0x44, 0x00,
  0x44, 0x44, 0x44, 0x28, 0x10, 0x10, 0x10, 0x00,
  0x78, 0x08, 0x10, 0x20, 0x40, 0x40, 0x78, 0x00, //90,
  0x38, 0x20, 0x20, 0x20, 0x20, 0x20, 0x38, 0x00,
  0x00, 0x40, 0x20, 0x10, 0x08, 0x04, 0x00, 0x00,
  0x38, 0x08, 0x08, 0x08, 0x08, 0x08, 0x38, 0x00,
  0x10, 0x28, 0x44, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC,
  0x30, 0x30, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x38, 0x04, 0x3C, 0x44, 0x3C, 0x00,
  0x40, 0x40, 0x78, 0x44, 0x44, 0x44, 0x78, 0x00,
  0x00, 0x00, 0x38, 0x44, 0x40, 0x44, 0x38, 0x00, //99
  0x04, 0x04, 0x3C, 0x44, 0x44, 0x44, 0x3C, 0x00, //100,
  0x00, 0x00, 0x38, 0x44, 0x78, 0x40, 0x38, 0x00,
  0x18, 0x20, 0x20, 0x78, 0x20, 0x20, 0x20, 0x00,
  0x00, 0x00, 0x3C, 0x44, 0x44, 0x3C, 0x04, 0x38,
  0x40, 0x40, 0x70, 0x48, 0x48, 0x48, 0x48, 0x00,
  0x10, 0x00, 0x10, 0x10, 0x10, 0x10, 0x18, 0x00,
  0x08, 0x00, 0x18, 0x08, 0x08, 0x08, 0x48, 0x30,
  0x40, 0x40, 0x48, 0x50, 0x60, 0x50, 0x48, 0x00,
  0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x18, 0x00,
  0x00, 0x00, 0x68, 0x54, 0x54, 0x44, 0x44, 0x00,
  0x00, 0x00, 0x70, 0x48, 0x48, 0x48, 0x48, 0x00, //110,
  0x00, 0x00, 0x38, 0x44, 0x44, 0x44, 0x38, 0x00,
  0x00, 0x00, 0x78, 0x44, 0x44, 0x44, 0x78, 0x40,
  0x00, 0x00, 0x3C, 0x44, 0x44, 0x44, 0x3C, 0x04,
  0x00, 0x00, 0x58, 0x24, 0x20, 0x20, 0x70, 0x00,
  0x00, 0x00, 0x38, 0x40, 0x38, 0x04, 0x38, 0x00,
  0x00, 0x20, 0x78, 0x20, 0x20, 0x28, 0x10, 0x00,
  0x00, 0x00, 0x48, 0x48, 0x48, 0x58, 0x28, 0x00,
  0x00, 0x00, 0x44, 0x44, 0x44, 0x28, 0x10, 0x00,
  0x00, 0x00, 0x44, 0x44, 0x54, 0x7C, 0x28, 0x00,
  0x00, 0x00, 0x48, 0x48, 0x30, 0x48, 0x48, 0x00, //120,
  0x00, 0x00, 0x48, 0x48, 0x48, 0x38, 0x10, 0x60,
  0x00, 0x00, 0x78, 0x08, 0x30, 0x40, 0x78, 0x00,
  0x18, 0x20, 0x20, 0x60, 0x20, 0x20, 0x18, 0x00,
  0x10, 0x10, 0x10, 0x00, 0x10, 0x10, 0x10, 0x00,
  0x30, 0x08, 0x08, 0x0C, 0x08, 0x08, 0x30, 0x00,
  0x28, 0x50, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  //    0x10,0x38,0x6C,0x44,0x44,0x7C,0x00,0x00   //127 DEL (OLD)
  0x38, 0x7C, 0x7C, 0x7C, 0x7C, 0x7C, 0x38, 0x00 //127 LED (NEW)
};

uint16_t counter = 0;
uint16_t counter2 = 0;
int bytes_togo = 0;
int char_mag_factor = CHAR_MAG_FACTOR;
WiFiClient client;
pfodESP8266BufferedClient bufferedClient;
byte old_paddles = 0xff; //"impossible", will have to change upon first run, will be sent
byte new_paddles;
byte ICACHE_RAM_ATTR read_paddles (void);
void ICACHE_RAM_ATTR write_da ( uint16_t xval, uint16_t yval);
void ICACHE_RAM_ATTR setint (void);
void ICACHE_RAM_ATTR beam_on (void);
void ICACHE_RAM_ATTR beam_off (void);
uint16_t x = 0; //both of these will have additional commands to the chip superimposed on them
uint16_t y = 0;
char coord[4];
int ii = 0;
char q;
void inline  handler (void);
void getput(void);
volatile int beam_state = 0;
void do_calib(void);
uint16_t c_yield = 0;
byte tempvar;
int tp;
byte ts[5];
byte virgin_chip;

//eeprom stuff
#define OFFSET_THESERVER 0
#define OFFSET_PORT 20
#define MAGIC_OFFSET 60
#define MAGIC_INT 0x55ff

char the_server[20];
int the_port;

int shouldSaveConfig = 0;

//ringbuff stuff
int point_ctr = 0;
int rbuf_min = 0;
int rbuf_max = 0;
int dbg_ctr = 0;
#define RINGBUFSIZE 3000
uint8_t* ringbuf;
uint16_t rbwindex = 0;
uint16_t rbrindex = RINGBUFSIZE - 1;
int rcount = 0;

char tempchar_for_port[10];

void setup()
{

  system_update_cpu_freq ( 160 ) ;  //you need this
  Serial.begin(115200); //only used in debugging
  EEPROM.begin(EEPROM_USE);


  if (!eprom_valid())
  {
    Serial.println("Eprom Not Valid");
    virgin_chip = 1;
    clear_eeprom();
    the_port = 0;
    the_server[0] = '\0';
  }


  WiFiManagerParameter custom_server("the_server", "PDP8/data server IP", the_server, 15);
  if (the_port == 0)
  {
    tempchar_for_port[0] = 0;
  }
  else
  {
    sprintf(tempchar_for_port, "%d", the_port);
  };

  WiFiManagerParameter custom_port("tempchar_for_port", "PDP8/data server port", tempchar_for_port, 5);

  WiFiManager wifiManager;
  wifiManager.addParameter(&custom_server);
  wifiManager.addParameter(&custom_port);
  wifiManager.setSaveConfigCallback(saveConfigCallback);

  ringbuf = (uint8_t*)malloc (RINGBUFSIZE);

  pinMode(HC165_CE, OUTPUT) ;
  pinMode(HC165_CP, OUTPUT) ;
  pinMode(HC165_PL, OUTPUT) ;
  pinMode(HC165_NDATA, INPUT) ;
  pinMode(BEAMPIN, OUTPUT) ;
  pinMode(TLV5617_CS, OUTPUT);
  digitalWrite(TLV5617_CS, HIGH);
  digitalWrite(HC165_CE, HIGH);
  digitalWrite(BEAMPIN, LOW);
  delay(100);


  if (read_paddles() == CALIB) do_calib();

  if (read_paddles() == SHOW_SETUP)
  {
    wifiManager.resetSettings();
    clear_eeprom();
    the_port = 0;
    the_server[0] = '\0';
    virgin_chip = 1;
  }

  wifiManager.autoConnect(WIFI_AP_NET, WIFI_AP_PASS);
  Serial.println("Connected to wifi");
  Serial.println("\nStarting connection...");

  if (shouldSaveConfig == 1)

  {
    //boy, I made a lot of trouble for myself with port as an integer...to save a little space :)
    strcpy(tempchar_for_port, custom_port.getValue());
    the_port = atoi(tempchar_for_port);
    write_int_eeprom(OFFSET_PORT, the_port);
    strcpy(the_server, custom_server.getValue());
    write_string_eeprom(OFFSET_THESERVER, the_server);
    shouldSaveConfig = 0;
  }
  the_port = read_int_eeprom(OFFSET_PORT);
  read_string_eeprom(OFFSET_THESERVER, the_server, 17);
  tp = the_port;
  parseBytes(the_server, '.', ts, 4, 10);

  if (client.connect(ts, tp)) {
    Serial.println("connected");

    bufferedClient.connect(&client); // buffer this client
    bufferedClient.setTimeout(10) ;
    setint();
    yield();
  }
}

void ICACHE_RAM_ATTR loop()
{
  int q;
  if (beam_state == 0)
  {
    getput();
    delayMicroseconds(abs(2000 - (2 * rcount)));
  }
  yield();
  if (rcount >= 5)
  {
    q = getring();
    if (q == 0)
    {
      ii++;
    }
    else
    {
      ii = 0;
    }

    if (ii == 2)
    {
      ii = 0;
      for (int k = 0; k < 4; k++)
        coord[k] = ( getring() & 0x3F);
      x = (((coord[0] | (coord[1] << 6))) + 512) % 1024;
      y = (((coord[2] | (coord[3] << 6))) + 512) % 1024;
      point_ctr++;
      minmax();
      write_da ( x, y);
      yield();
      counter++;
      if (counter > LOOPS_TO_PADDLES) {
        counter = 0;
        new_paddles = read_paddles();
        if (new_paddles != old_paddles) {
          old_paddles = new_paddles;
          yield();
          bufferedClient.write(new_paddles);
          yield();
          Serial.println (new_paddles);
        }
      }
    }
  }
  else
  {
    counter2++;
    if (counter2 == 1)
    {
      yield();
      counter2 = 0;
    }
  }
}



byte ICACHE_RAM_ATTR read_paddles (void)
{
  byte c = 0;
  byte bitval;
  int i;

  //when pressed, pullups go to ground, but hardware compliments

  GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, HC165_CE_FAST);
  GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, HC165_PL_FAST);
  GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, HC165_PL_FAST);
  GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, HC165_CE_FAST);
  GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, HC165_CP_FAST);

  for ( i = 0; i < 8; i++)
  {
    bitval = digitalRead(HC165_NDATA);
    c |= (bitval << (7 - i));
    //delay(0);
    //clock to get another
    GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, HC165_CP_FAST);
    GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, HC165_CP_FAST);
  }
  GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, HC165_CE_FAST);

  yield();
  return c;
}
void  ICACHE_RAM_ATTR write_da ( uint16_t xval, uint16_t yval)
{
  uint16_t intval;
  int i;
  yield();
  //perform spi/i2c here to write x, then write y and latch both upon entry of y
  xval = xval << 2;
  yval = yval << 2;

  xval = (xval | 0xC000); //do second
  yval = (yval | 0x5000); //do first

  GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, TLV5617_CS_FAST);
  GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, TLV5617_CLK_FAST);
  GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, TLV5617_CS_FAST);
  GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, TLV5617_CLK_FAST);

  for ( i = 0; i < 16; i++)
  {
    intval = (yval << i) & 0x8000;
    if (intval > 0)
    {

      GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, TLV5617_DTA_FAST);
    }
    else
    {
      GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, TLV5617_DTA_FAST);
    }
    GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, TLV5617_CLK_FAST);
    GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, TLV5617_CLK_FAST);
  }
  GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, TLV5617_CS_FAST);
  GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, TLV5617_CLK_FAST);
  GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, TLV5617_CS_FAST);
  GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, TLV5617_CLK_FAST);

  yield();

  for ( i = 0; i < 16; i++)
  {
    intval = (xval << i) & 0x8000;
    if (intval > 0)
    {
      GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, TLV5617_DTA_FAST);
    }
    else
    {

      GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, TLV5617_DTA_FAST);
    }
    GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, TLV5617_CLK_FAST);
    GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, TLV5617_CLK_FAST);
  }
  beam_off;
  GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, TLV5617_CS_FAST);
  delayMicroseconds(4); //THIS IS FOR SETTLING
  beam_on();
  yield();
}

void ICACHE_RAM_ATTR beam_on (void)
{
  if (beam_state != 1)
  {
    timer0_write(ESP.getCycleCount() + BEAM_TIMELENGTH);
    GPIO_REG_WRITE(GPIO_OUT_W1TS_ADDRESS, BEAMPIN_FAST);
    beam_state = 1;
  }

}
void inline  handler (void)
{
  GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, BEAMPIN_FAST);
  timer0_write(ESP.getCycleCount() + BEAM_TIMELENGTH);
  beam_state = 0;
}


void ICACHE_RAM_ATTR beam_off (void)
{
  while (beam_state == 1)yield();
  GPIO_REG_WRITE(GPIO_OUT_W1TC_ADDRESS, BEAMPIN_FAST);
  beam_state = 0;
}

void ICACHE_FLASH_ATTR  do_calib(void)

{
  int z;
  uint16_t l_x;
  uint16_t l_y;
  char  tempchar[20];
  unsigned long mymillis;
  int r;

  setint();
  mymillis = millis() + 8000;
  Serial.print ("X");
  Serial.println ("X");
  Serial.println("Calib");
  Serial.println("calib");
  Serial.println("calib");
  yield();
  while (read_paddles() != BOOTIT)
  {
    r = read_paddles();

    if (z != r)
    {
      sprintf(tempchar, "%d", r); // converts to decimal base.
      Serial.println();
      Serial.println (r);
      Serial.println(tempchar);
      z = r;
    }



    put_a_line(tempchar, 512, 512);
    if ( mymillis > millis())put_a_line ("SLOB", 750, 80);
    put_a_line(THE_VERSION, 140, 80);

    //Serial.print(".");
    yield();
    //top line
    l_y = 0;
    for (l_x = 0;  l_x < 1024; l_x += 4)
    {
      write_da(l_x, l_y);
    }

    l_y = 1023;
    for (l_x = 0;  l_x < 1024; l_x += 4)
    {
      write_da(l_x, l_y);
    }

    l_x = 0;
    for (l_y = 0;  l_y < 1024; l_y += 4)
    {
      write_da(l_x, l_y);
    }

    l_x = 1023;
    for (l_y = 0;  l_y < 1024; l_y += 4)
    {
      write_da(l_x, l_y);
    }
  }
}

void setint ()
{
  noInterrupts();
  timer0_isr_init();
  timer0_attachInterrupt(handler);
  timer0_write(ESP.getCycleCount() + BEAM_TIMELENGTH);
  interrupts();

}

void parseBytes(const char* str, char sep, byte* bytes, int maxBytes, int base) {
  for (int i = 0; i < maxBytes; i++) {
    bytes[i] = strtoul(str, NULL, base);  // Convert byte
    str = strchr(str, sep);               // Find next separator
    if (str == NULL || *str == '\0') {
      break;                            // No more separators, exit
    }
    str++;                                // Point to next character after separator
  }
}

inline bool ringspace()
{
  return (rcount < RINGBUFSIZE);
}

inline uint16_t ringavail()
{
  return rcount;
}

bool putring (uint8_t b)
{

  *(ringbuf + rbwindex) = b;
  if (++rbwindex == RINGBUFSIZE)
  {
    rbwindex = 0;
  }
  rcount++;
}

byte getring ()
{
  if (++rbrindex == RINGBUFSIZE)
  {
    rbrindex = 0;
  }
  rcount--;
  return *(ringbuf + rbrindex);
}

void getput (void)
{
  int  x, b;
  static int qq = 0;
  yield();
  //fill up ring buffer

  if ( rcount < RINGBUFSIZE)
  {
    b = bufferedClient.available();


    if (b < ((RINGBUFSIZE - 1) - rcount) && b > 5)
    {
      for (x = 0; x < b; x++)
      {
        putring(bufferedClient.read());
        yield();
      }

    }
  }
}



void minmax()
//used for debugging ring buffer
{

  if (rcount > rbuf_max)
  {
    rbuf_max = rcount;
  }
  if (rcount < rbuf_max)
  {
    rbuf_min = rcount;
  }
  dbg_ctr++;
  if (dbg_ctr >= 3000)
  {
    Serial.print ("rbuf_min=");
    Serial.println (rbuf_min);
    Serial.print ("rbuf_max=");
    Serial.println (rbuf_max);
    Serial.print ("point_ctr=");
    Serial.println (point_ctr);
    dbg_ctr = 0;
    rbuf_min = 0;
    rbuf_max = 0;
    point_ctr = 0;

  }

}

void put_a_char(char thechar, int the_x, int the_y)
{
  byte ch;
  int a;
  int thelines;
  int o;
  unsigned int off;

  //Serial.print (thechar);

  for (thelines = 0; thelines < 8; thelines++)
  {
    off = ((byte(thechar) - 32) * 8) + (thelines);
    ch = fdata[off];
    for (a = 0; a < 8; a++)
    {
      o = (ch << a) & 128;
      if (o != 0)
      {
        write_da(the_x + (a << CHAR_MAG_FACTOR), (the_y - (thelines << CHAR_MAG_FACTOR)));
      }

    }


  }
}


void put_a_line (char * thestring, int the_x, int the_y)
{
  int sl;
  int c;
  int lx, ly;


  sl = strlen( thestring);
  lx = the_x - ((((sl - 1) * 8) << CHAR_MAG_FACTOR)  >> 1);
  ly = the_y + (((8 << CHAR_MAG_FACTOR)) >> 1);
  // Serial.print ("thestring=");
  //Serial.println (thestring);

  for (c = 0; c < sl; c++)
  {
    put_a_char( thestring[c], lx + ((c * 8) << CHAR_MAG_FACTOR), ly);
  }
}


void clear_eeprom(void)
//this also puts down a magic byte so that a "virgin" chip goes into AP mode.
{
  int i;
  for (i = 0; i < EEPROM_USE; i++)
  {
    EEPROM.write(i, 0);
  }
  write_int_eeprom( MAGIC_OFFSET, MAGIC_INT);
  EEPROM.commit();
  delay(100);
}

void write_int_eeprom(unsigned int the_offset, unsigned int the_int)
{


  byte lByte = ((the_int) & 0xff);
  byte hByte = ((the_int >> 8) & 0xff);
  EEPROM.write (the_offset, lByte);
  EEPROM.write (the_offset + 1, hByte);
  EEPROM.commit();
  delay(100);

}


int read_int_eeprom (int the_offset)
{


  byte lByte = EEPROM.read(the_offset);
  byte hByte = EEPROM.read (the_offset + 1);
  return ((lByte << 0) & 0xff) + ((hByte << 8) & 0xff00);
}

bool eprom_valid()
{
  bool retval;


  retval = (read_int_eeprom(MAGIC_OFFSET) == MAGIC_INT);

  if (retval)
  {
    Serial.println("Eprom is valid");
  }
  else
  {
    Serial.println("Eprom is not valid");
  }
  return retval;
}


void read_string_eeprom(unsigned int the_offset, char* thestring, int the_len)
{
  int done = 0;
  for ( x = 0; (x <= the_len && !done); x++)
  {
    thestring[x] = EEPROM.read(the_offset + x);
    if (thestring[x] == 0) done = 1;
  }

}

void write_string_eeprom (unsigned int the_offset, char* thestring)
{
  int sl;
  sl = strlen(thestring);

  for (int x = 0; x < sl; x++)
  {
    EEPROM.write (the_offset + x, thestring[x]);
  }
  EEPROM.write (the_offset + sl + 1, 0);
  EEPROM.commit();
  delay(200);

}

void die(void)//used in debugging
{
  while (1 == 1)
  {
    yield;
  }
}

void saveConfigCallback () {
  //Serial.println("Should save config");
  shouldSaveConfig = 1;
}

